from dataclasses import dataclass
from datetime import datetime

import pytz

from cloud_governance.common.clouds.cloudability.templates.cloudability_dimensions import COST_USAGE_REPORT_DIMENSIONS, \
    COST_USAGE_REPORT_METRICS


@dataclass
class CostUsageReportData:
    ResourceName: str
    OfferingType: str
    UsageFamily: str
    InstanceCategory: str
    UsageType: str
    ItemDescription: str
    Region: str
    ServiceName: str
    StartDate: str
    CostCenter: str
    AccountName: str
    AccountId: str
    PublicCloudName: str
    Cost: float
    timestamp: datetime
    ResourcePlanId: str
    IndexId: str = ""
    ReportGenerate: str = ""

    def __init__(self, report_generated_type: str, **kwargs):
        for k, v in kwargs.items():
            if k in COST_USAGE_REPORT_DIMENSIONS:
                key = COST_USAGE_REPORT_DIMENSIONS.get(k)
            elif k in COST_USAGE_REPORT_METRICS:
                key = COST_USAGE_REPORT_METRICS.get(k)
                v = float(v)
            else:
                key = k
            setattr(self, key, v)
        self.IndexId = f'{report_generated_type}-{self.StartDate}-{self.ResourceName}-{self.AccountId}'
        self.timestamp = datetime.strptime(self.StartDate, '%Y-%m-%d')
        self.ReportGenerate = report_generated_type

    def to_dict(self):
        return {key: value for key, value in self.__dict__.items()}



import json
import os
import time
from datetime import timedelta, datetime

import boto3

from cloud_governance.common.logger.init_logger import logger


class CloudTrailOperations:
    SEARCH_SECONDS = 10
    SLEEP_SECONDS = 120
    LOOKBACK_DAYS = 30

    def __init__(self, region_name: str):
        self.__cloudtrail = boto3.client('cloudtrail', region_name=region_name)
        self.__global_cloudtrail = boto3.client('cloudtrail', region_name='us-east-1')
        self.__iam_client = boto3.client('iam')

    def __check_filter_username(self, username: str, event: dict):
        """
        This method checks for the exact username.
        @param username:
        @param event:
        @return:
        """
        if '-' in username:
            event_record = json.loads(event.get('CloudTrailEvent'))
            if event_record.get('userIdentity').get('type') == "IAMUser":
                role_arn = event_record.get('userIdentity').get('arn')
                assumerole_username, event = self.__get_username_by_role(role_arn, "CreateUser", "AWS::IAM::User")
                if assumerole_username:
                    username = assumerole_username
        return username

    def __get_cloudtrail_responses(self, start_time: datetime, end_time: datetime, resource_arn: str):
        """
        This method return all the responses from the cloudtrail for a certain time period
        @param start_time:
        @param end_time:
        @param resource_arn:
        @return:
        """
        responses = []
        if start_time:
            response = self.__global_cloudtrail.lookup_events(StartTime=start_time, EndTime=end_time, LookupAttributes=[{'AttributeKey': 'ResourceName', 'AttributeValue': resource_arn}])
        else:
            response = self.__global_cloudtrail.lookup_events(LookupAttributes=[{'AttributeKey': 'ResourceName', 'AttributeValue': resource_arn}])
        responses.extend(response.get('Events'))
        while response.get('NextToken'):
            if start_time:
                response = self.__global_cloudtrail.lookup_events(StartTime=start_time, EndTime=end_time, LookupAttributes=[{'AttributeKey': 'ResourceName', 'AttributeValue': resource_arn}], NextToken=response.get('NextToken'))
            else:
                response = self.__global_cloudtrail.lookup_events(LookupAttributes=[{'AttributeKey': 'ResourceName', 'AttributeValue': resource_arn}], NextToken=response.get('NextToken'))
            responses.extend(response.get('Events'))
        return responses

    def __get_role_start_time_end_time(self, role_arn: str, event_name: str):
        """
        This method get the start_time, end_time from the role/user.
        @param role_arn:
        @param event_name:
        @return:
        """
        try:
            role_name = role_arn.split('/')[-1]
            if event_name == 'CreateUser':
                start_time = self.__iam_client.get_user(UserName=role_name)['User'].get('CreateDate')
            elif event_name == 'CreateRole':
                start_time = self.__iam_client.get_role(RoleName=role_name)['Role'].get('CreateDate')
            end_time = start_time + timedelta(seconds=self.SEARCH_SECONDS)
            return [start_time, end_time]
        except:
            return ['', '']

    def __get_username_by_role(self, resource_arn: str, event_name: str, resource_type: str):
        """
        This method get the username from the role
        @param resource_arn:
        @param event_name:
        @param resource_type:
        @return:
        """
        try:
            start_time, end_time = self.__get_role_start_time_end_time(role_arn=resource_arn, event_name=event_name)
            responses = self.__get_cloudtrail_responses(start_time, end_time, resource_arn)
            for event in responses:
                if event.get('EventName') == event_name:
                    if event.get('Resources'):
                        for resource in event.get('Resources'):
                            if resource.get('ResourceType') == resource_type:
                                if resource.get('ResourceName') == resource_arn:
                                    role_username = event.get('Username')
                                    role_username = self.__check_filter_username(role_username, event)
                                    username, assumed_event = self.__check_event_is_assumed_role(event.get('CloudTrailEvent'))
                                    if username:
                                        return [username, assumed_event]
                                    return [role_username, event]
            return ['', '']
        except Exception as err:
            return ['', '']

    def __ger_username_from_arn(self, resource_arn: str):
        """
        This method return the username from the userIdentity arn
        """
        events = self.__get_cloudtrail_responses(resource_arn=resource_arn, start_time=None, end_time=None)
        for event in events:
            for resource in event.get('Resources'):
                if resource.get('ResourceType') == 'AWS::STS::AssumedRole':
                    if resource.get('ResourceName') == resource_arn:
                        username, assumed_event = self.__check_event_is_assumed_role(event.get('CloudTrailEvent'))
                        if username:
                            return [username, assumed_event]
                        return [event.get('Username'), event]
        return ['', '']

    def __extract_username_from_arn(self, arn: str, user_type: str):
        """
        Extract username from ARN based on userIdentity type.
        @param arn: The ARN from userIdentity
        @param user_type: The userIdentity type (IAMUser, AssumedRole, FederatedUser, etc.)
        @return: username or empty string
        """
        if not arn or '/' not in arn:
            return ''

        # ARN formats:
        # IAMUser: arn:aws:iam::account:user/[path/]username
        # AssumedRole: arn:aws:sts::account:assumed-role/role-name/session-name
        # FederatedUser: arn:aws:sts::account:federated-user/username
        # Root: arn:aws:iam::account:root (no slash, return 'root')

        parts = arn.split('/')
        if len(parts) < 2:
            # No username in ARN (e.g., root user)
            return parts[-1] if parts else ''

        # Last part is always the username/session-name
        return parts[-1]

    def __check_event_is_assumed_role(self, cloudtrail_event_str: str):
        """
        This method extracts username from userIdentity ARN for IAM users and AssumedRole users.
        For SAML SSO (AssumedRole), it extracts the username from the session name in the ARN.
        For IAM users, it extracts the username from the ARN path.
        @param cloudtrail_event_str: JSON string of CloudTrailEvent
        @return: [username, parsed_event] or [False, '']
        """
        try:
            cloudtrail_event = json.loads(cloudtrail_event_str)
            user_identity = cloudtrail_event.get('userIdentity', {})
            user_type = user_identity.get('type')
            arn = user_identity.get('arn')

            # Handle supported user types by extracting from ARN
            if user_type in ('AssumedRole', 'IAMUser', 'FederatedUser'):
                username = self.__extract_username_from_arn(arn, user_type)
                if username:
                    # Return parsed event wrapped in a dict with CloudTrailEvent key for consistency
                    return [username, {'CloudTrailEvent': cloudtrail_event_str}]

            # For Root or other types without proper ARN
            if user_type == 'Root':
                return ['root', {'CloudTrailEvent': cloudtrail_event_str}]

            return [False, '']
        except Exception as err:
            return [False, '']

    def get_full_responses(self, **kwargs):
        """
        This method return all responses
        @param kwargs:
        @return:
        """
        responses = []
        response = self.__cloudtrail.lookup_events(**kwargs)
        responses.extend(response['Events'])
        while response.get('NextToken'):
            response = self.__cloudtrail.lookup_events(**kwargs, NextToken=response.get('NextToken'))
            responses.extend(response['Events'])
        if responses:
            return responses
        return []

    def __get_user_by_resource_id(self, start_time: datetime, end_time: datetime, resource_id: str, resource_type: str, event_type: str):
        """
        This method find the username of the resource_id with given resource_type
        @param start_time:
        @param end_time:
        @param resource_id:
        @param resource_type:
        @return:
        """
        try:
            responses = self.get_full_responses(StartTime=start_time, EndTime=end_time, LookupAttributes=[{
                'AttributeKey': event_type, 'AttributeValue': resource_type}])
            for event in responses:
                if event.get('EventName') == resource_type:
                    if event.get('Resources'):
                        for resource in event.get('Resources'):
                            if resource.get('ResourceName') == resource_id:
                                username, assumed_event = self.__check_event_is_assumed_role(event.get('CloudTrailEvent'))
                                if username:
                                    return [username, assumed_event]
                                return [event.get('Username'), event]
                if event.get('Resources'):
                    for resource in event.get('Resources'):
                        if resource.get('ResourceType') == resource_type:
                            if resource.get('ResourceName') == resource_id:
                                username, assumed_event = self.__check_event_is_assumed_role(event.get('CloudTrailEvent'))
                                if username:
                                    return [username, assumed_event]
                                return [event.get('Username'), event]
            return ['', '']
        except Exception as err:
            return ['', '']

    def __get_time_difference(self, start_time: datetime):
        """
        This method returns seconds in difference of current time and start time
        @param start_time:
        @return:
        """
        current_time = datetime.now(start_time.tzinfo).replace(tzinfo=None)
        diff = (current_time - start_time.replace(tzinfo=None))
        return (diff.days * 24 * 60 * 60) + diff.seconds

    def get_username_by_instance_id_and_time(self, resource_id: str, resource_type: str, start_time: datetime = '', event_type: str = 'ResourceType', end_time: datetime = None):
        """
        This method find Username in cloud trail events according to start_time and resource_id
        @param event_type:
        @param start_time:
        @param resource_id:
        @param resource_type:
        @param end_time:
        @return: if user not found it return empty string

        """
        if start_time and not end_time:
            delay_seconds = int(os.environ.get('SLEEP_SECONDS', self.SLEEP_SECONDS))
            if self.__get_time_difference(start_time=start_time) <= delay_seconds:
                time.sleep(delay_seconds)
            search_time = timedelta(seconds=self.SEARCH_SECONDS)
            end_time = start_time + search_time
            start_time = start_time - search_time
        else:
            if not start_time and not end_time:
                start_time = datetime.now() - timedelta(days=self.LOOKBACK_DAYS)
                end_time = datetime.now()
        username, event = self.__get_user_by_resource_id(start_time, end_time, resource_id, resource_type, event_type)
        return self.__check_filter_username(username, event)

    def get_stop_time(self, resource_id: str, event_name: str):
        """
        This method return the time of when instance is stopped
        @param resource_id:
        @param event_name:
        @return:
        """
        responses = []
        try:
            response = self.__cloudtrail.lookup_events(LookupAttributes=[
                {'AttributeKey': 'ResourceName', 'AttributeValue': resource_id},
            ])
            responses.extend(response['Events'])
            while response.get('NextToken'):
                response = self.__cloudtrail.lookup_events(LookupAttributes=[
                    {'AttributeKey': 'ResourceName', 'AttributeValue': resource_id},
                ], NextToken=response.get('NextToken'))
                responses.extend(response['Events'])

            for event in responses:
                if event.get('EventName') == event_name:
                    if event.get('Resources'):
                        for resource in event.get('Resources'):
                            if resource.get('ResourceName') == resource_id:
                                return event.get('EventTime')
            return ''
        except:
            return ''

    def set_cloudtrail(self, region_name: str):
        self.__cloudtrail = boto3.client('cloudtrail', region_name=region_name)

    def get_last_time_accessed(self, resource_id: str, event_name: str, start_time: datetime, end_time: datetime, **kwargs):
        """
        This method return last accesses time
        @param resource_id:
        @param event_name:
        @param start_time:
        @param end_time:
        @return:
        """
        try:
            events = self.__cloudtrail.lookup_events(StartTime=start_time, EndTime=end_time, LookupAttributes=[{
                'AttributeKey': 'ResourceName', 'AttributeValue': resource_id
            }])['Events']
            if events:
                events = sorted(events, key=lambda event: event['EventTime'], reverse=True)
                events = [event for event in events if event.get('EventName') not in ('CreateTags', 'DeleteTags')]
                if events:
                    if events[0].get('EventName') == event_name:
                        return events[0].get('EventTime')
                if kwargs:
                    if len(events) == 1:
                        if events[0].get('EventName') == kwargs['optional_event_name'][0]:
                            return events[0].get('EventTime')
                    elif len(events) == 2:
                        if events[0].get('EventName') in kwargs['optional_event_name'] and events[1].get('EventName') in kwargs['optional_event_name']:
                            return events[0].get('EventTime')
            else:
                return start_time - timedelta(days=4)
        except Exception as err:
            logger.info(f'{err}')
            return None
        return None



from cloud_governance.common.logger.init_logger import logger
from cloud_governance.policy.policy_operations.aws.tag_cluster.remove_cluster_tags import RemoveClusterTags
from cloud_governance.policy.policy_operations.aws.tag_cluster.tag_cluster_resouces import TagClusterResources


def tag_cluster_resource(cluster_name: str = '', mandatory_tags: dict = None, region: str = 'us-east-2', tag_operation: str = 'yes', cluster_only: bool = False):
    """
    This method scan for cluster name in all the cluster resources
    :return: list of cluster resources according to cluster name
    """

    if tag_operation == "update":
        action = 'Tag'
        dry_run = 'no'
    else:
        action = 'read'
        dry_run = 'yes'
    tag_cluster_resources = TagClusterResources(cluster_prefix=["kubernetes.io/cluster", "sigs.k8s.io/cluster-api-provider-aws/cluster"], cluster_name=cluster_name,
                                                input_tags=mandatory_tags, region=region, dry_run=dry_run, cluster_only=cluster_only)

    func_resource_list = [tag_cluster_resources.cluster_instance,
                          tag_cluster_resources.cluster_volume,
                          tag_cluster_resources.cluster_ami,
                          tag_cluster_resources.cluster_snapshot,
                          tag_cluster_resources.cluster_network_interface,
                          tag_cluster_resources.cluster_load_balancer,
                          tag_cluster_resources.cluster_load_balancer_v2,
                          tag_cluster_resources.cluster_dhcp_option,
                          tag_cluster_resources.cluster_subnet,
                          tag_cluster_resources.cluster_route_table,
                          tag_cluster_resources.cluster_vpc_endpoint,
                          tag_cluster_resources.cluster_nat_gateway,
                          tag_cluster_resources.cluster_internet_gateway,
                          tag_cluster_resources.cluster_security_group,
                          tag_cluster_resources.cluster_elastic_ip,
                          tag_cluster_resources.cluster_vpc,
                          tag_cluster_resources.cluster_role,
                          tag_cluster_resources.cluster_user,
                          tag_cluster_resources.cluster_s3_bucket,
                          ]
    if cluster_only:
        logger.info(f"{action} {len(func_resource_list)} cluster resources for cluster name '{cluster_name}' in region {region}:")
    else:
        logger.info(f"{action} {len(func_resource_list)} cluster resources for cluster name '{cluster_name}' in region {region}:")
        logger.info(f"{action} 4 non-cluster resources in region {region}:")
    if not cluster_name:
        func_resource_list[0]()
        func_resource_list = func_resource_list[1:-3]
    else:
        func_resource_list[0]()
        func_resource_list = func_resource_list[1:]
    for _, func in enumerate(func_resource_list):
        func()


def remove_cluster_resources_tags(region: str, cluster_name: str, input_tags: dict, cluster_only: bool = False):
    """
    This method removes the tags from the AWS resources
    @param cluster_only:
    @param region:
    @param cluster_name:
    @param input_tags:
    @return:
    """
    remove_cluster_tags = RemoveClusterTags(region=region, cluster_name=cluster_name, cluster_prefix='kubernetes.io/cluster/', input_tags=input_tags, cluster_only=cluster_only)
    func_resource_list = [remove_cluster_tags.cluster_instance,
                          remove_cluster_tags.cluster_volume,
                          remove_cluster_tags.cluster_images,
                          remove_cluster_tags.cluster_snapshot,
                          remove_cluster_tags.cluster_load_balancer,
                          remove_cluster_tags.cluster_load_balancer_v2,
                          remove_cluster_tags.cluster_network_interface,
                          remove_cluster_tags.cluster_elastic_ip,
                          remove_cluster_tags.cluster_security_group,
                          remove_cluster_tags.cluster_vpc,
                          remove_cluster_tags.cluster_subnet,
                          remove_cluster_tags.cluster_route_table,
                          remove_cluster_tags.cluster_internet_gateway,
                          remove_cluster_tags.cluster_dhcp_option,
                          remove_cluster_tags.cluster_vpc_endpoint,
                          remove_cluster_tags.cluster_nat_gateway,
                          remove_cluster_tags.cluster_network_acl,
                          remove_cluster_tags.cluster_role,
                          remove_cluster_tags.cluster_user,
                          remove_cluster_tags.cluster_s3_bucket]
    if cluster_name:
        logger.info(f"{len(func_resource_list)} remove tag cluster resources from '{cluster_name}' in region {region}:")
    else:
        logger.info(f"{len(func_resource_list)} remove tag cluster resources  in region {region}:")
    response = func_resource_list[0]()
    for func in func_resource_list[1:]:
        func(instance_tags=response)
    logger.info(response)



import logging

from ibm_cloud_sdk_core.authenticators import IAMAuthenticator

from cloud_governance.main.environment_variables import environment_variables


class IBMAuthenticator:
    """
    Refer: https://cloud.ibm.com/apidocs/vpc/latest
    Refer: https://github.com/IBM/ibm-cloud-sdk-common?tab=readme-ov-file
    """

    def __init__(self):
        logging.disable(logging.DEBUG)
        self.env_config = environment_variables
        if hasattr(self.env_config, 'IBM_ACCOUNT_ID'):
            self.account_id = self.env_config.IBM_ACCOUNT_ID
        else:
            self.account_id = self.env_config.environment_variables_dict.get('IBM_ACCOUNT_ID')
        if hasattr(self.env_config, 'IBM_CLOUD_API_KEY'):
            self.__api_key = self.env_config.IBM_CLOUD_API_KEY
        else:
            self.__api_key = self.env_config.environment_variables_dict.get('IBM_CLOUD_API_KEY')
        self.iam_authenticator = IAMAuthenticator(self.__api_key)



import boto3

from cloud_governance.common.clouds.aws.iam.iam_operations import IAMOperations
from cloud_governance.common.clouds.aws.utils.utils import Utils
from cloud_governance.common.logger.init_logger import logger


class RemoveUserTags:
    """
    This class contain methods of removing user tags
    """

    def __init__(self, remove_keys: list, username: str = ''):
        self.remove_keys = remove_keys
        self.username = username
        self.iam_client = boto3.client('iam')
        self.IAMOperations = IAMOperations()
        self.get_detail_resource_list = Utils().get_details_resource_list

    def __cluster_user(self, tags: list):
        """
        This method check the user is cluster or not
        @param tags:
        @return:
        """
        for tag in tags:
            if 'kubernetes.io/cluster' in tag.get('Key'):
                return True
        return False

    def user_tags_remove(self):
        """
        This method check the user
        @return:
        """
        users = self.get_detail_resource_list(func_name=self.iam_client.list_users, input_tag='Users',
                                              check_tag='Marker')
        count = 0
        for user in users:
            user_name = user.get('UserName')
            if self.username:
                if self.username == user_name:
                    self.iam_client.untag_user(UserName=user_name, TagKeys=self.remove_keys)
                    count += 1
                    logger.info(f'Username :: {user_name} :: {self.remove_keys}')
                    break
            else:
                user_tags = self.IAMOperations.get_user_tags(username=user_name)
                if not self.__cluster_user(tags=user_tags):
                    self.iam_client.untag_user(UserName=user_name, TagKeys=self.remove_keys)
                    count += 1
                    logger.info(f'Username :: {user_name} :: {self.remove_keys}')
        return count



from abc import ABC, abstractmethod


class AbstractTaggingOperations(ABC):
    """
    This class is abstract tagging operations to all the clouds
    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_resources_list(self, tag_name: str, tag_value: str = ''):
        raise NotImplementedError()

    @abstractmethod
    def tag_resources_list(self, resources_list: list, update_tags_dict: dict):
        raise NotImplementedError()



from datetime import datetime, timezone, date, timedelta
import time

from cloud_governance.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from cloud_governance.common.elasticsearch.elastic_upload import ElasticUpload
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.main.environment_variables import environment_variables


class YearlySavingsReport:
    """
    This class aggregates yearly savings and per-month savings for the current year
    from policy execution data and uploads to a dedicated ES index
    Uses month-by-month queries and deduplication to avoid counting resources multiple times
    """

    NAT_GATEWAY_HOURLY_COST = 0.045
    ELASTIC_IP_HOURLY_COST = 0.005

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__es_host = self.__environment_variables_dict.get('es_host', '')
        self.__es_port = self.__environment_variables_dict.get('es_port', '')
        self.__elastic_operations = ElasticSearchOperations(es_host=self.__es_host, es_port=self.__es_port) if self.__es_host else None
        self.__elastic_upload = ElasticUpload()
        self.__policy_es_index = self.__environment_variables_dict.get('es_index', 'cloud-governance-policy-es-index')
        self.__yearly_savings_es_index = self.__environment_variables_dict.get('yearly_savings_es_index')
        account = self.__environment_variables_dict.get('account', 'PERF-DEPT')
        self.__account = account.upper().replace('OPENSHIFT-', '').replace('OPENSHIFT', '').strip()
        # Check for custom date range from environment variables. This won't upload to ES.
        self.__custom_start_date = self.__environment_variables_dict.get('yearly_savings_start_date', '')
        self.__custom_end_date = self.__environment_variables_dict.get('yearly_savings_end_date', '')

    def __get_last_day_of_month(self, year: int, month: int):
        """
        Get the last day of a month, handling leap years
        @param year: Year
        @param month: Month (1-12)
        @return: Last day of month
        """
        if month == 2:
            # Handle February (leap year)
            if year % 4 == 0 and (year % 100 != 0 or year % 400 == 0):
                return 29
            else:
                return 28
        elif month in [4, 6, 9, 11]:
            return 30
        else:
            return 31

    def __get_savings_value(self, captured_date: date, policy_name: str):
        """
        @param captured_date: date object
        @param policy_name: policy name
        @return: calculated savings value
        """
        savings = 0
        end_of_year = date(captured_date.year, 12, 31)
        remaining_days = (end_of_year - captured_date).days

        if policy_name == 'unused_nat_gateway':
            savings = remaining_days * 24 * self.NAT_GATEWAY_HOURLY_COST
        elif policy_name == 'ip_unattached':
            savings = remaining_days * 24 * self.ELASTIC_IP_HOURLY_COST

        return savings

    def __process_monthly_query(self, month_start: str, month_end: str):
        """
        Process a single monthly query and return resource-level results
        @param month_start: Start date string (YYYY-MM-DD)
        @param month_end: End date string (YYYY-MM-DD)
        @return: Dictionary of {resource_id: {captured_date, policy_name, savings}}
        """
        if not self.__elastic_operations:
            return {}

        query = {
            "size": 0,
            "query": {
                "bool": {
                    "must": [
                        {
                            "term": {
                                "PublicCloud.keyword": {
                                    "value": "AWS"
                                }
                            }
                        },
                        {
                            "term": {
                                "account.keyword": {
                                    "value": self.__account
                                }
                            }
                        }
                    ],
                    "must_not": [
                        {
                            "terms": {
                                "policy.keyword": [
                                    "zombie_cluster_resource", "instance_run", "ebs_in_use",
                                    "s3_inactive", "optimize_resources_report", "instance_idle",
                                    "cluster_run", "skipped_resources", "ec2_idle", "empty_roles",
                                    "unused_access_key"
                                ]
                            }
                        }
                    ],
                    "filter": [{
                        "range": {
                            "timestamp": {
                                "gte": month_start,
                                "lte": month_end,
                                "format": "yyyy-MM-dd"
                            }
                        }
                    }]
                }
            },
            "aggs": {
                "PolicyName": {
                    "terms": {
                        "field": "policy.keyword",
                        "size": 20,
                        "order": {"_key": "desc"}
                    },
                    "aggs": {
                        "CapturedDate": {
                            "terms": {
                                "field": "timestamp",
                                "size": 10000
                            },
                            "aggs": {
                                "ResourceId": {
                                    "terms": {
                                        "field": "ResourceId.keyword",
                                        "size": 10000,
                                        "order": {"_key": "desc"}
                                    },
                                    "aggs": {
                                        "Savings": {
                                            "max": {
                                                "field": "TotalYearlySavings",
                                                "missing": 0
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        try:
            response = self.__elastic_operations.post_query(
                query=query,
                es_index=self.__policy_es_index,
                result_agg=True
            )

            if not response or 'PolicyName' not in response:
                logger.warning(f'No response or unexpected structure for {month_start} to {month_end}')
                return {}

            monthly_resources = {}
            policy_buckets = response.get('PolicyName', {}).get('buckets', [])

            logger.debug(f'DEBUG: Found {len(policy_buckets)} policies in query results')
            for policy in policy_buckets:
                logger.debug(f'DEBUG: Policy found: {policy.get("key")} with {len(policy.get("CapturedDate", {}).get("buckets", []))} captured dates')

            for policy in policy_buckets:
                policy_name = policy.get('key')

                for es_capture_date_values in policy.get('CapturedDate', {}).get('buckets', []):
                    captured_date_str = es_capture_date_values.get('key_as_string', '')

                    if captured_date_str:
                        captured_date = datetime.strptime(captured_date_str[:10], "%Y-%m-%d").date()
                    else:
                        captured_date = datetime.now(timezone.utc).date()

                    for resource in es_capture_date_values.get('ResourceId', {}).get('buckets', []):
                        resource_id = resource.get('key')
                        savings = resource.get('Savings', {}).get('value', 0)

                        if resource_id in monthly_resources:
                            if monthly_resources[resource_id]['captured_date'] > captured_date:
                                monthly_resources[resource_id]['captured_date'] = captured_date
                                if savings == 0:
                                    savings = self.__get_savings_value(captured_date, policy_name)
                                if monthly_resources[resource_id]['savings'] > savings:
                                    monthly_resources[resource_id]['savings'] = savings
                        else:
                            if savings == 0:
                                savings = self.__get_savings_value(captured_date, policy_name)
                            monthly_resources[resource_id] = {
                                'captured_date': captured_date,
                                'policy_name': policy_name,
                                'savings': savings
                            }

            policy_summary = {}
            for resource_id, resource_data in monthly_resources.items():
                policy_name = resource_data.get('policy_name')
                savings = resource_data.get('savings', 0)
                policy_summary[policy_name] = policy_summary.get(policy_name, 0) + savings

            logger.info(f'Monthly resources summary by policy: {policy_summary}')
            logger.info(f"Found {len(monthly_resources)} resources for {month_start} to {month_end}")
            return monthly_resources

        except Exception as err:
            logger.error(f'Error processing month {month_start} to {month_end}: {err}')
            return {}

    def __calculate_month_savings(self, year: int, month: int, start_day: int, end_day: int):
        """
        Calculate savings for a specific month
        @param year: Year
        @param month: Month (1-12)
        @param start_day: Start day of month
        @param end_day: End day of month
        @return: Dictionary of {policy_name: savings_value}
        """
        month_start = date(year, month, start_day).strftime("%Y-%m-%d")
        month_end = date(year, month, end_day).strftime("%Y-%m-%d")

        logger.info(f'Calculating savings for {month_start} to {month_end}')

        return self.__get_yearly_savings(start_date=month_start, end_date=month_end)

    def __get_yearly_savings(self, start_date: str = None, end_date: str = None):
        """
        This method returns the yearly savings of the policies.
        Queries month-by-month to avoid Elasticsearch overload.
        @param start_date: Start date string (YYYY-MM-DD). If None, defaults to Jan 1 of current year.
        @param end_date: End date string (YYYY-MM-DD). If None, defaults to Dec 31 of current year.
        @return: Dictionary of {policy_name: total_savings}
        """
        if not start_date:
            start_date = datetime.now(timezone.utc).date().replace(day=1, month=1).strftime("%Y-%m-%d")
        if not end_date:
            end_date = datetime.now(timezone.utc).date().replace(day=31, month=12).strftime("%Y-%m-%d")

        logger.info(f"Getting yearly savings from {start_date} to {end_date}")

        monthly_ranges = self.__split_date_range_by_month(start_date, end_date)
        logger.debug(f"Split into {len(monthly_ranges)} monthly queries")

        all_resources = {}

        for i, (month_start, month_end) in enumerate(monthly_ranges, 1):
            logger.info(f"Processing month {i}/{len(monthly_ranges)}: {month_start} to {month_end}")
            monthly_resources = self.__process_monthly_query(month_start, month_end)

            for resource_id, resource_data in monthly_resources.items():
                if resource_id in all_resources:
                    if all_resources[resource_id]['captured_date'] > resource_data['captured_date']:
                        all_resources[resource_id]['captured_date'] = resource_data['captured_date']
                        if all_resources[resource_id]['savings'] > resource_data['savings']:
                            all_resources[resource_id]['savings'] = resource_data['savings']
                else:
                    all_resources[resource_id] = resource_data

            if i < len(monthly_ranges):
                time.sleep(0.5)

        result = self.__get_total_policy_sum(all_resources)
        return result

    def __split_date_range_by_month(self, start_date: str, end_date: str):
        """
        Split date range into monthly chunks
        @param start_date: Start date string (YYYY-MM-DD)
        @param end_date: End date string (YYYY-MM-DD)
        @return: List of (month_start, month_end) tuples as strings
        """
        if not start_date or not end_date:
            raise ValueError(f"Both start_date and end_date must be provided. Got: start_date={start_date}, end_date={end_date}")

        start = datetime.strptime(start_date, "%Y-%m-%d").date()
        end = datetime.strptime(end_date, "%Y-%m-%d").date()

        if start > end:
            raise ValueError(f"start_date ({start_date}) must be <= end_date ({end_date})")

        monthly_ranges = []
        current_start = start

        while current_start <= end:
            if current_start.month == 12:
                current_end = date(current_start.year + 1, 1, 1) - timedelta(days=1)
            else:
                current_end = date(current_start.year, current_start.month + 1, 1) - timedelta(days=1)
            if current_end > end:
                current_end = end
            monthly_ranges.append((
                current_start.strftime("%Y-%m-%d"),
                current_end.strftime("%Y-%m-%d")
            ))
            if current_end.month == 12:
                current_start = date(current_end.year + 1, 1, 1)
            else:
                current_start = date(current_end.year, current_end.month + 1, 1)

        return monthly_ranges

    def __get_total_policy_sum(self, all_resources: dict):
        """
        Calculate total savings by policy
        @param all_resources: dict of {resource_id: {savings, policy_name, captured_date}}
        @return: dict of {policy_name: total_savings}
        """
        savings_result = {}

        for resource_id, values in all_resources.items():
            policy_name = values.get('policy_name')
            savings = values.get('savings', 0)
            savings_result[policy_name] = round(savings_result.get(policy_name, 0) + savings, 3)

        return savings_result

    def __update_yearly_savings(self, year: int, all_months_data: dict, total_annual_saving: float):
        """
        Update yearly savings in Elasticsearch with all months data.
        Creates a new document if year doesn't exist, or updates existing one.
        @param year: Year (e.g., 2026)
        @param all_months_data: Dictionary of {month_number: savings_value} for all months
        @param total_annual_saving: Total cumulative savings for the year
        @return: True if successful
        """
        try:
            current_date = datetime.now(timezone.utc).date()
            year_id = f"{year}-{self.__account}"

            data = {
                'year': year,
                'total_saving': round(total_annual_saving, 3),
                'last_updated': current_date.isoformat(),
                'timestamp': datetime.now(timezone.utc),
                'policy': 'yearly_savings_report',
                'index_id': year_id,
                'account': self.__account
            }

            for month_num in range(1, 13):
                data[f'month_{month_num}'] = round(all_months_data.get(month_num, 0), 3)

            try:
                if self.__elastic_operations.verify_elastic_index_doc_id(index=self.__yearly_savings_es_index, doc_id=year_id):
                    self.__elastic_operations.update_elasticsearch_index(
                        index=self.__yearly_savings_es_index,
                        id=year_id,
                        metadata=data
                    )
                    logger.info(f"Updated yearly savings for year {year}")
                else:
                    self.__elastic_operations.upload_to_elasticsearch(
                        index=self.__yearly_savings_es_index,
                        data=data,
                        id=year_id
                    )
                    logger.info(f"Created yearly savings document for year {year}")
            except Exception as e:
                logger.warning(f"Update check failed, trying create: {e}")
                try:
                    self.__elastic_operations.upload_to_elasticsearch(
                        index=self.__yearly_savings_es_index,
                        data=data,
                        id=year_id
                    )
                    logger.info(f"Created yearly savings document for year {year}")
                except Exception as create_err:
                    logger.error(f"Failed to create document: {create_err}")
                    raise create_err

            return True

        except Exception as err:
            logger.error(f"Error updating yearly savings: {err}")
            raise err

    @logger_time_stamp
    def run(self, start_date: str = None, end_date: str = None):
        """
        Main method to run the yearly savings report

        @param start_date: Optional start date string (YYYY-MM-DD). If None, checks environment variable, then defaults to Jan 1 of current year.
        @param end_date: Optional end date string (YYYY-MM-DD). If None, checks environment variable, then defaults to today.
        @return: dict with summary
        """
        if not start_date:
            start_date = self.__custom_start_date
        if not end_date:
            end_date = self.__custom_end_date

        if start_date and end_date:
            try:
                logger.info(f'Using custom date range: {start_date} to {end_date}')

                month_savings = self.__get_yearly_savings(start_date=start_date, end_date=end_date)
                total_savings = sum(month_savings.values()) if month_savings else 0.0

                logger.info(f'Custom date range - Policy savings: {month_savings}')
                logger.info(f'Custom date range - Total savings: ${total_savings:,.2f}')

                return {
                    'status': 'success',
                    'custom_date_range': True,
                    'start_date': start_date,
                    'end_date': end_date,
                    'policy_savings': month_savings,
                    'total_yearly_savings': total_savings
                }
            except ValueError as e:
                logger.error(f'Invalid date format: {e}. Expected YYYY-MM-DD')
                return {'status': 'error', 'message': f'Invalid date format: {e}'}

        current_date = datetime.now(timezone.utc)
        current_year = current_date.year
        current_month = current_date.month
        today = current_date.date()

        logger.info(f'Using default date range: current year {current_year}')

        all_months_data = {}

        for month in range(1, 13):
            try:
                if month < current_month:
                    end_day = self.__get_last_day_of_month(current_year, month)
                    month_savings = self.__calculate_month_savings(
                        year=current_year,
                        month=month,
                        start_day=1,
                        end_day=end_day
                    )
                    all_months_data[month] = sum(month_savings.values()) if month_savings else 0.0

                elif month == current_month:
                    month_savings = self.__calculate_month_savings(
                        year=current_year,
                        month=month,
                        start_day=1,
                        end_day=today.day
                    )
                    all_months_data[month] = sum(month_savings.values()) if month_savings else 0.0

                else:
                    all_months_data[month] = 0.0

                if month < 12:
                    time.sleep(0.2)

            except Exception as e:
                logger.warning(f"Error calculating savings for month {month}: {e}, setting to 0")
                all_months_data[month] = 0.0

        total_annual_saving = sum(all_months_data.values())

        logger.info(f'Monthly savings: {all_months_data}')
        logger.info(f'Total annual savings: ${total_annual_saving:,.2f}')

        if self.__elastic_operations:
            try:
                self.__update_yearly_savings(
                    year=current_year,
                    all_months_data=all_months_data,
                    total_annual_saving=total_annual_saving
                )
                logger.info(f'Successfully uploaded yearly savings to {self.__yearly_savings_es_index}')
                return {
                    'status': 'success',
                    'records_uploaded': 1,
                    'year': current_year,
                    'total_yearly_savings': total_annual_saving,
                    'monthly_savings': all_months_data
                }
            except Exception as err:
                logger.error(f'Error uploading to ES: {err}')
                return {'status': 'error', 'message': str(err)}
        else:
            logger.warning('ES not configured')
            return {'status': 'no_upload', 'message': 'ES not configured'}



from datetime import datetime

import typeguard

from cloud_governance.common.clouds.aws.utils.common_methods import get_tag_value_from_tags
from cloud_governance.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from cloud_governance.main.environment_variables import environment_variables
from cloud_governance.policy.policy_operations.aws.zombie_cluster.zombie_cluster_common_methods import \
    ZombieClusterCommonMethods
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.policy.aws.zombie_cluster_resource import ZombieClusterResources


@typeguard.typechecked
def __get_resource_list(region, delete: bool = False, resource: str = '', cluster_tag: str = '',
                        resource_name: str = '', service_type: str = ' '):
    cluster_prefix = environment_variables.environment_variables_dict.get('CLUSTER_PREFIX')
    zombie_cluster_resources = ZombieClusterResources(cluster_prefix=cluster_prefix, delete=delete,
                                                      region=region, cluster_tag=cluster_tag,
                                                      resource_name=resource_name)
    zombie_cluster_resources_dict = {'zombie_cluster_volume': zombie_cluster_resources.zombie_cluster_volume,
                                     'zombie_cluster_ami': zombie_cluster_resources.zombie_cluster_ami,
                                     'zombie_cluster_snapshot': zombie_cluster_resources.zombie_cluster_snapshot,
                                     'zombie_cluster_security_group': zombie_cluster_resources.zombie_cluster_security_group,
                                     'zombie_cluster_elastic_ip': zombie_cluster_resources.zombie_cluster_elastic_ip,
                                     'zombie_cluster_network_interface': zombie_cluster_resources.zombie_cluster_network_interface,
                                     'zombie_cluster_load_balancer': zombie_cluster_resources.zombie_cluster_load_balancer,
                                     'zombie_cluster_load_balancer_v2': zombie_cluster_resources.zombie_cluster_load_balancer_v2,
                                     'zombie_cluster_vpc': zombie_cluster_resources.zombie_cluster_vpc,
                                     'zombie_cluster_subnet': zombie_cluster_resources.zombie_cluster_subnet,
                                     'zombie_cluster_route_table': zombie_cluster_resources.zombie_cluster_route_table,
                                     'zombie_cluster_internet_gateway': zombie_cluster_resources.zombie_cluster_internet_gateway,
                                     'zombie_cluster_dhcp_option': zombie_cluster_resources.zombie_cluster_dhcp_option,
                                     'zombie_cluster_vpc_endpoint': zombie_cluster_resources.zombie_cluster_vpc_endpoint,
                                     'zombie_cluster_nat_gateway': zombie_cluster_resources.zombie_cluster_nat_gateway,
                                     'zombie_cluster_network_acl': zombie_cluster_resources.zombie_cluster_network_acl,
                                     'zombie_cluster_role': zombie_cluster_resources.zombie_cluster_role,
                                     'zombie_cluster_user': zombie_cluster_resources.zombie_cluster_user,
                                     'zombie_cluster_s3_bucket': zombie_cluster_resources.zombie_cluster_s3_bucket}
    ec2_zombie_resource_services = [zombie_cluster_resources.zombie_cluster_ami,
                                    zombie_cluster_resources.zombie_cluster_volume,
                                    zombie_cluster_resources.zombie_cluster_load_balancer,
                                    zombie_cluster_resources.zombie_cluster_load_balancer_v2,
                                    zombie_cluster_resources.zombie_cluster_snapshot,
                                    zombie_cluster_resources.zombie_cluster_vpc_endpoint,
                                    zombie_cluster_resources.zombie_cluster_dhcp_option,
                                    zombie_cluster_resources.zombie_cluster_route_table,
                                    zombie_cluster_resources.zombie_cluster_security_group,
                                    zombie_cluster_resources.zombie_cluster_nat_gateway,
                                    zombie_cluster_resources.zombie_cluster_network_acl,
                                    zombie_cluster_resources.zombie_cluster_network_interface,
                                    zombie_cluster_resources.zombie_cluster_elastic_ip,
                                    zombie_cluster_resources.zombie_cluster_internet_gateway,
                                    zombie_cluster_resources.zombie_cluster_subnet,
                                    zombie_cluster_resources.zombie_cluster_vpc
                                    ]
    iam_zombie_resource_services = [zombie_cluster_resources.zombie_cluster_role]
    # user_delete permission is very problematic operation, it might affect other users.
    # [zombie_cluster_resources.zombie_cluster_user]
    s3_zombie_resource_services = [zombie_cluster_resources.zombie_cluster_s3_bucket]
    scan_func_resource_list = []
    delete_func_resource_list = []
    if resource:
        scan_func_resource_list = [zombie_cluster_resources_dict[resource]]
        delete_func_resource_list = [zombie_cluster_resources_dict[resource]]
    else:
        if service_type:
            if service_type == 'ec2_zombie_resource_service':
                scan_func_resource_list.extend(ec2_zombie_resource_services)
                delete_func_resource_list.extend(ec2_zombie_resource_services)
            elif service_type == 'iam_zombie_resource_service':
                scan_func_resource_list.extend(iam_zombie_resource_services)
                delete_func_resource_list.extend(iam_zombie_resource_services)
            elif service_type == 's3_zombie_resource_service':
                scan_func_resource_list.extend(s3_zombie_resource_services)
                delete_func_resource_list.extend(s3_zombie_resource_services)
        else:
            scan_func_resource_list = ec2_zombie_resource_services + iam_zombie_resource_services + s3_zombie_resource_services
            delete_func_resource_list = ec2_zombie_resource_services + iam_zombie_resource_services + s3_zombie_resource_services
    if delete:
        return delete_func_resource_list
    else:
        return scan_func_resource_list


@typeguard.typechecked
@logger_time_stamp
def zombie_cluster_resource(delete: bool = False, region: str = 'us-east-2', resource: str = '', cluster_tag: str = '',
                            resource_name: str = '', service_type: str = ''):
    """
    This method returns zombie cluster resources,
    How its works? if not exist an instance cluster, the resource is zombie
    if delete true it will delete the zombie resource
    :return: list of zombie resources
    """
    zombie_result = {}
    all_cluster_data = []

    if delete:
        action = 'Delete'
        if resource:
            func_resource_list = __get_resource_list(region, delete, resource, cluster_tag, resource_name, service_type)
        else:
            func_resource_list = __get_resource_list(region, delete, resource, cluster_tag, resource_name, service_type)
    else:
        action = 'Scan'
        if resource:
            func_resource_list = __get_resource_list(region, delete, resource, cluster_tag, resource_name, service_type)
        elif resource_name:
            func_resource_list = __get_resource_list(region, delete, resource, cluster_tag, resource_name, service_type)
        else:
            func_resource_list = __get_resource_list(region, delete, resource, cluster_tag, resource_name, service_type)
    logger.info(f'{action} {len(func_resource_list)} cluster zombies resources in region {region}:')
    notify_data = {}
    delete_data = {}
    cluster_data = {}
    zombie_cluster_common_methods = ZombieClusterCommonMethods(region=region)
    cluster_delete_days = {}
    zombie_cluster_resources_ids = {}
    zombie_cluster_resources_data = {}
    for func in func_resource_list:
        resource_data, cluster_left_out_days = func()
        if resource_data:
            notify_data, delete_data, cluster_data = zombie_cluster_common_methods.collect_notify_cluster_data(
                resource_data=resource_data,
                cluster_left_out_days=cluster_left_out_days,
                notify_data=notify_data, delete_data=delete_data, cluster_data=cluster_data, func_name=func.__name__)
            resource_data_list = sorted(set(resource_data.values()))
            for resource_id, cluster_name in resource_data.items():
                zombie_cluster_resources_ids.setdefault(cluster_name.strip(), []).append(resource_id)
                cluster_delete_days[cluster_name] = get_tag_value_from_tags(tag_name='ClusterDeleteDays',
                                                                            tags=cluster_left_out_days.get(cluster_name,
                                                                                                           []))
                zombie_cluster_resources_data.setdefault(cluster_name, []).append(func.__name__)
            resource_data_list = [func.__name__.replace("zombie_cluster_", "") + ':' + item for item in
                                  resource_data_list]
            logger.info(f'key: {func.__name__}, count: {len(resource_data)}, data: {resource_data_list}')
            zombie_result[func.__name__] = {'count': len(resource_data), 'data': resource_data_list}
            all_cluster_data.extend(resource_data_list)
    zombie_cluster_common_methods.send_mails_to_cluster_user(notify_data=notify_data, delete_data=delete_data,
                                                             cluster_data=cluster_data)
    zombie_cluster_result = {}
    for cluster_data in all_cluster_data:
        _, cluster_name = cluster_data.split(':', 1)
        zombie_cluster_result[cluster_name] = zombie_cluster_result.get(cluster_name, 0) + 1
    zombie_cluster_result = [{'ResourceId': key, 'ZombieClusterTag': key, 'ClusterResourcesCount': value}
                             for key, value in zombie_cluster_result.items()]
    es_operations = ElasticSearchOperations()
    if es_operations.check_elastic_search_connection():
        environment_variables_dict = environment_variables.environment_variables_dict
        es_index = environment_variables_dict.get('es_index')
        account = environment_variables_dict.get('account', '')
        if zombie_cluster_result:
            regional_zombie_cluster_ids = []
            global_zombie_cluster_resources = ['zombie_cluster_role', 'zombie_cluster_s3_bucket']
            for zombie_cluster_id, resources_data in zombie_cluster_resources_data.items():
                if (len(resources_data) > 0 and resources_data[0] in global_zombie_cluster_resources) \
                        or (len(resources_data) > 1 and resources_data[1] in global_zombie_cluster_resources):
                    if region == 'us-east-1':
                        regional_zombie_cluster_ids.append(zombie_cluster_id)
                    else:
                        continue
                else:
                    regional_zombie_cluster_ids.append(zombie_cluster_id)
            for zombie_cluster in zombie_cluster_result:
                if zombie_cluster['ZombieClusterTag'] in regional_zombie_cluster_ids:
                    zombie_cluster['region_name'] = region
                    zombie_cluster['account'] = account
                    zombie_cluster['ResourceIds'] = zombie_cluster_resources_ids[zombie_cluster['ResourceId']]
                    zombie_cluster['CleanUpDays'] = cluster_delete_days[zombie_cluster['ZombieClusterTag']]
                    es_operations.upload_to_elasticsearch(data=zombie_cluster.copy(), index=es_index)
                    logger.info(f'Uploaded the policy results to elasticsearch index: {es_index}')
        else:
            logger.info(f'No data to upload on @{account}  at {datetime.utcnow()}')
    else:
        logger.error('ElasticSearch host is not pingable, Please check ')
    return zombie_result



import json

import boto3

from cloud_governance.common.clouds.aws.support.support_operations import SupportOperations
from cloud_governance.common.clouds.aws.utils.common_methods import get_tag_value_from_tags
from cloud_governance.common.logger.init_logger import logger

# @Todo, focusing only on cost-optimizing service: Need to find the tags for, rds, route53, ecr


class OptimizeResourcesReport:
    COST_OPIMIZING_REPORTS = {
        'ec2_reports': ['Amazon EC2 instances over-provisioned for Microsoft SQL Server',
                        'Low Utilization Amazon EC2 Instances',
                        'Amazon EC2 instances consolidation for Microsoft SQL Server',
                        'Amazon EC2 Reserved Instance Lease Expiration', 'Amazon EC2 Instances Stopped'],
        'ebs_reports': ['Amazon EBS over-provisioned volumes', 'Underutilized Amazon EBS Volumes'],
        'eip_reports': ['Unassociated Elastic IP Addresses'],
        's3_reports': ['Amazon S3 Bucket Lifecycle Policy Configured',
                       'Amazon S3 version-enabled buckets without lifecycle policies configured'],
        'rds_reports': ['Amazon RDS Idle DB Instances'],
        'load_balancer_reports': ['Idle Load Balancers'],
        'lambda_reports': ['AWS Lambda Functions with Excessive Timeouts', 'AWS Lambda Functions with High Error Rates',
                           'AWS Lambda over-provisioned functions for memory size'],
        'ecr_reports': ['Amazon ECR Repository Without Lifecycle Policy Configured'],
        'route_53_reports': ['Amazon Route 53 Latency Resource Record Sets', 'Amazon Route 53 Name Server Delegations']
    }

    def __init__(self):
        self.__support_operations = SupportOperations()

    def __get_tags(self, name: str, region_name: str, resource_id: str):
        """
        This method returns the aws client
        :param name:
        :return:
        """
        ec2_reports = self.COST_OPIMIZING_REPORTS['ec2_reports']
        ebs_reports = self.COST_OPIMIZING_REPORTS['ebs_reports']
        eip_reports = self.COST_OPIMIZING_REPORTS['eip_reports']
        s3_reports = self.COST_OPIMIZING_REPORTS['s3_reports']
        load_balancer_reports = self.COST_OPIMIZING_REPORTS['load_balancer_reports']

        def lower_data(data: list):
            return list(map(lambda item: item.lower(), data))
        try:
            if name.lower() in lower_data(ec2_reports) + lower_data(ebs_reports) + lower_data(eip_reports):
                return boto3.client('ec2', region_name=region_name). \
                    describe_tags(Filters=[{'Name': 'resource-id', 'Values': [resource_id]}]).get('Tags', [])
            elif name.lower() in lower_data(s3_reports):
               return boto3.client('s3', region_name=region_name).get_bucket_tagging(Bucket=resource_id).get('TagSet', [])
            elif name.lower() in lower_data(load_balancer_reports):
                tags = boto3.client('elb', region_name=region_name).\
                    describe_tags(LoadBalancerNames=[resource_id]).get('TagDescriptions', [])
                for tag in tags:
                    if tag.get('LoadBalancerName') == resource_id:
                        return tag.get('Tags')
                tags = boto3.client('elbv2', region_name=region_name). \
                    describe_tags(ResourceArns=[resource_id]).get('TagDescriptions', [])
                for tag in tags:
                    if tag.get('ResourceArn') == resource_id:
                        return tag.get('Tags')
        except Exception as err:
            logger.error(err)
        return []

    def __get_optimization_reports(self):
        """
        This method returns the report data
        :return:
        :rtype:
        """
        report_list = self.__support_operations.get_trusted_advisor_reports()
        optimize_resource_list = []
        unique_report = set()
        for report_name, resources in report_list.items():
            if report_name:
                for key, values in resources.items():
                    name = values.get('metadata', {}).get('name')
                    unique_report.add(name)
                    flagged_resources_list = values.get('reports', {}).get('flaggedResources')
                    if flagged_resources_list:
                        resource_values = [item.replace(' ', '') for item in
                                           values.get('metadata', {}).get('metadata', [])]
                        for flagged_resources in flagged_resources_list:
                            resource_location = ''
                            if report_name == 'cost_optimizing':
                                resource_location = 1
                            resources = {}
                            for idx, item in enumerate(flagged_resources.get('metadata', [])):
                                if resource_values[idx] in [f'Day{i}' for i in range(1, 15)]:
                                    continue
                                resources[resource_values[idx]] = str(item)
                                if resource_location and resource_location == idx:
                                    tags = self.__get_tags(name, region_name=flagged_resources.get('region'), resource_id=item)
                                    user = get_tag_value_from_tags(tags=tags, tag_name='User')
                                    resources['User'] = user
                                    resources['ResourceId'] = item
                            resources['ReportName'] = name
                            resources['Report'] = report_name
                            if 'LastUpdatedTime' in resources:
                                del resources['LastUpdatedTime']
                            optimize_resource_list.append(resources)
        return optimize_resource_list

    def run(self):
        """
        This method start the report collection
        :return:
        :rtype:
        """
        return self.__get_optimization_reports()



import json
from urllib.parse import urlparse, parse_qs

from ibm_platform_services import ResourceControllerV2

from cloud_governance.common.clouds.ibm.account.ibm_authenticator import IBMAuthenticator


class PlatformServiceOperations(IBMAuthenticator):

    def __init__(self):
        super().__init__()
        self.__client = ResourceControllerV2(authenticator=self.iam_authenticator)

    def get_resource_instances(self):
        """
        This method returns all the service instances
        :return:
        """
        responses = self.__client.list_resource_instances().get_result()
        resources = responses['resources']
        while responses.get('next_url'):
            parsed_url = urlparse(responses['next_url'])
            params = parse_qs(parsed_url.query)
            if params and 'start' in params:
                start = params['start'][0]
                responses = self.__client.list_resource_instances(start=start).get_result()
                resources.extend(responses['resources'])
        resource_instances = []
        for resource in resources:
            if resource['type'] == 'resource_instance':
                resource_instances.append(resource)
        return {'global': resource_instances}



import json
import os
from datetime import datetime, timedelta
from ast import  literal_eval

from typeguard import typechecked

from cloud_governance.common.clouds.gcp.google_account import GoogleAccount
from cloud_governance.common.elasticsearch.elastic_upload import ElasticUpload
from cloud_governance.common.google_drive.upload_to_gsheet import UploadToGsheet
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.main.environment_variables import environment_variables
from cloud_governance.common.logger.init_logger import logger


class CostBillingReports:
    """
    This class is responsible for generation cost billing report for Budget, Actual, Forecast
    """

    DEFAULT_YEARS = 12
    DEFAULT_ROUND_DIGITS = 3

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__database_name = self.__environment_variables_dict.get('GCP_DATABASE_NAME', '')
        self.__table_name = self.__environment_variables_dict.get('GCP_DATABASE_TABLE_NAME', '')
        self.__gcp_account = GoogleAccount()
        self.__gsheet_id = self.__environment_variables_dict.get('SPREADSHEET_ID')
        self.__cloud_name = self.__environment_variables_dict.get('PUBLIC_CLOUD_NAME').upper()
        self.update_to_gsheet = UploadToGsheet()
        self.elastic_upload = ElasticUpload()

    @logger_time_stamp
    def __next_twelve_months(self):
        """
        This method returns the next 12 month, year
        :return:
        """
        months = 12
        year = datetime.now().year
        next_month = datetime.now().month + 1
        month_year = []
        for idx in range(months):
            month = str((idx+next_month) % months)
            c_year = year
            if len(month) == 1:
                month = f'0{month}'
            if month == '00':
                month = 12
                year = year+1
            month_year.append((str(month), c_year))
        return month_year

    @typechecked()
    @logger_time_stamp
    def __prepare_usage_query(self, first_year_month: str = None, second_year_month: str = None):
        """
        This method prepare the query for usage
        :param first_year_month: #YYYYMM
        :param second_year_month: #YYYYMM
        :return:
        """
        if not first_year_month and not second_year_month:
            current_month = datetime.now().replace(day=1)
            past_month = current_month - timedelta(days=1)
            first_year_month = past_month.strftime("%Y%m")
            second_year_month = current_month.strftime("%Y%m")
        logger.info(f'StartMonth: {first_year_month}, EndMonth: {second_year_month}')
        fetch_monthly_invoice_query = f"""
                SELECT ifnull(project.ancestors[SAFE_OFFSET(1)].display_name, 'NA') as folder_name,
                ifnull(project.ancestry_numbers, 'NA') as folder_id, invoice.month, ifnull(project.id, 'GCP-refund/credit') as project_name, ifnull(project.number, '000000000000') as project_id,
                (SUM(CAST(cost AS NUMERIC)) + SUM(IFNULL((SELECT SUM(CAST(c.amount AS NUMERIC))
                FROM UNNEST(credits) AS c), 0))) AS total_cost
                FROM `{self.__database_name}.{self.__table_name}`
                where  invoice.month BETWEEN '{first_year_month}' AND '{second_year_month}'
                GROUP BY 1, 2, 3, 4, 5
                ORDER BY 3
                """
        fetch_monthly_folders_query = f"""
                SELECT TO_JSON_STRING(project.ancestors) as project_folders, project.number, invoice.month, ifnull(project.ancestry_numbers, 'NA') as folder_id
                FROM `{self.__database_name}.{self.__table_name}`
                where  invoice.month BETWEEN '{first_year_month}' AND '{second_year_month}' GROUP BY 1, 2, 3, 4 ORDER BY invoice.month
                """
        return [fetch_monthly_invoice_query, fetch_monthly_folders_query]

    @typechecked()
    @logger_time_stamp
    def __organized_results(self, data_rows: list):
        """
        This method organize the results to be uploaded to elastic search
        :param data_rows:
        :return:
        """
        compress_gcp_data = {}  # compress data based on budget_id
        for row in data_rows:
            month = row.get('Month')
            cost_center, allocated_budget, years, owner = 0, 0, '', 'Others'
            project_budget_account_id = 0
            budget_approved_account_name = ''
            for idx, _id in enumerate((row.get('folder_ids')+[row.get('ProjectId')])[::-1]):  # start from reverse [root, sub_child, child]
                cost_center, allocated_budget, years, owner = self.update_to_gsheet.get_cost_center_budget_details(account_id=_id, dir_path='/tmp')
                if cost_center > 0:
                    project_budget_account_id = _id
                    budget_approved_account_name = row.get(_id)
                    break
            parent_index = len(row.get("folder_ids"))
            index = f'{project_budget_account_id}-{budget_approved_account_name}-{month}'
            if index in compress_gcp_data:
                compress_gcp_data[index]['Actual'] = round(compress_gcp_data[index]['Actual'] + row.get('Actual'), 3)
                compress_gcp_data[index]['Projects'].append({
                    'Project': row.get('Project'),
                    'Actual': round(row.get('Actual'), self.DEFAULT_ROUND_DIGITS),
                    'ProjectId': row.get('ProjectId')
                })
                if parent_index > compress_gcp_data[index]['total_folders']:
                    compress_gcp_data[index]['Account'] = row.get(f'parent{parent_index}', 'NA')
                    compress_gcp_data[index]['AccountId'] = row.get(f'parent{parent_index}_id', 'NA')
            else:
                project_cost_data = {'CloudName': self.__cloud_name, 'CostCenter': cost_center, 'Owner': owner,
                                     'Budget': round(allocated_budget / self.DEFAULT_YEARS, self.DEFAULT_ROUND_DIGITS),
                                     'Forecast': 0,
                                     'AllocatedBudget': round(allocated_budget, self.DEFAULT_ROUND_DIGITS),
                                     'BudgetId': project_budget_account_id,
                                     'Account': row.get(f'parent{parent_index}', 'NA'),
                                     'AccountId': row.get(f'parent{parent_index}_id', 'NA'),
                                     'Actual': round(row.get('Actual'), self.DEFAULT_ROUND_DIGITS),
                                     'filter_date': row.get('filter_date'), 'Month': row.get('Month'),
                                     'start_date': row.get('start_date'), 'timestamp': row.get('timestamp'),
                                     'Projects': [{'Project': row.get('Project'),
                                                   'Actual': round(row.get('Actual'), self.DEFAULT_ROUND_DIGITS),
                                                   'ProjectId': row.get('ProjectId')}],
                                     'index_id': f"{row.get('start_date')}-{project_budget_account_id}-{row.get(f'parent{parent_index}', 'NA').lower()}",
                                     'total_folders': parent_index}
                if budget_approved_account_name:
                    project_cost_data['Account'] = (f"{row.get(f'parent{parent_index}', 'NA')}/"
                                                    f"{budget_approved_account_name}")
                    project_cost_data['AccountId'] = f"{row.get(f'parent{parent_index}_id', 'NA')}/{project_budget_account_id}"
                compress_gcp_data[index] = project_cost_data
        return self.__second_layer_filter(items=list(compress_gcp_data.values()))

    @typechecked()
    @logger_time_stamp
    def __second_layer_filter(self, items: list):
        """
        This method aggregates the results which have the same Account name
        :param items:
        :return:
        """
        filtered_result = {}
        for item in items:
            account = item.get('Account')
            month = item.get('Month')
            index = f'{account}-{month}'
            if index in filtered_result:
                filtered_result[index]['Budget'] += item.get('Budget')
                filtered_result[index]['Actual'] += item.get('Actual')
                if item.get('BudgetId') != filtered_result[index]['BudgetId']:
                    filtered_result[index]['AllocatedBudget'] += item.get('AllocatedBudget')
                filtered_result[index]['Projects'].extend(item.get('Projects'))
            else:
                filtered_result[index] = item
        return list(filtered_result.values())

    # @Todo Add forecast values in future
    @typechecked()
    @logger_time_stamp
    def __forecast_for_next_months(self, cost_data: list):
        """
        This method returns the forecast of next twelve months data
        :param cost_data:
        :return:
        """
        forecast_cost_data = []
        month_years = self.__next_twelve_months()
        month = (datetime.now().month - 1) % 12
        if month == 0:
            month = 12
        if len(str(month)) == 1:
            month = f'0{month}'
        year = datetime.now().year
        cache_start_date = f'{year}-{str(month)}-01'
        for data in cost_data:
            if cache_start_date == data.get('start_date') and data.get('CostCenter') > 0:
                for m_y in month_years:
                    m, y = m_y[0], m_y[1]
                    start_date = f'{y}-{m}-01'
                    timestamp = datetime.strptime(start_date, "%Y-%m-%d")
                    index_id = f'{start_date}-{data.get("Account").lower()}'
                    month = datetime.strftime(timestamp, "%Y %b")
                    projects = []
                    for project in data.get('Projects'):
                        project['Actual'] = 0
                        projects.append(project)
                    forecast_cost_data.append({
                        **data,
                        'Actual': 0,
                        'start_date': start_date,
                        'timestamp': timestamp,
                        'index_id': index_id,
                        'Projects': projects,
                        'filter_date': f'{start_date}-{month.split()[-1]}',
                        'Month': month}
                    )
        return forecast_cost_data

    @typechecked()
    @logger_time_stamp
    def __get_aggregated_folder_details(self, query_data: list):
        """
        This method gives the unique folder_names from the data
        :param query_data:
        :return:
        """
        project_folders = {}
        for data in query_data:
            index = f'{data.get("number")}'
            month = data.get('month')
            project_folder_id = data.get('folder_id')
            insert_data = False
            if index not in project_folders:
                insert_data = True
            else:
                insert_data = month >= project_folders.get(index).get('month')
            if insert_data:
                updated_data = {'month': month, 'folder_id': project_folder_id}
                for folders in literal_eval(data.get('project_folders')):
                    folder_id = folders.get('resource_name').split('/')[-1]
                    folder_name = folders.get('display_name')
                    updated_data[folder_id] = folder_name
                project_folders[index] = updated_data
        return project_folders

    @typechecked()
    @logger_time_stamp
    def __get_parent_folders(self, folder_ids: list, folders_data: dict, project_id: str):
        """
        This method returns the list of parent folders of Project
        :param folder_ids:
        :param folders_data:
        :param project_id:
        :return:
        """
        parent_folders = {}
        for idx, _id in enumerate(folder_ids):
            parent_folders.update({
                f'parent{idx + 1}': folders_data[project_id].get(_id),
                f'parent{idx + 1}_id': _id,
                f'{_id}': folders_data[project_id].get(_id),
            })
        return parent_folders

    @logger_time_stamp
    def __get_big_query_data(self):
        """
        This method collect the data from the big query and filter the data
        :return:
        """
        cost_usage_queries = self.__prepare_usage_query()
        query_rows = self.__gcp_account.query_list(cost_usage_queries)
        folders_data = self.__get_aggregated_folder_details(query_rows[1])
        agg_data = {}
        for cst_row in query_rows[0]:
            project_id, bill_month, total_cost = cst_row.get('project_id').strip(), cst_row.get('month'), float(cst_row.get('total_cost'))
            folder_ids = folders_data.get(project_id).get('folder_id').split('/')[2:-1] if folders_data.get(project_id) else cst_row.get('folder_id').split('/')[2:-1]
            folder_name = cst_row.get('folder_name')
            index = f"{project_id}-{bill_month}"
            parents_folders = self.__get_parent_folders(folder_ids, folders_data, project_id) if project_id in folders_data else {}
            if agg_data.get(index):
                total_cost = float(cst_row.get('total_cost')) + agg_data[index]['Actual']
            agg_data[index] = {
                'folder_name': parents_folders.get(f'parent{len(folder_ids)}', 'NA'),
                'start_date': f'{bill_month[:4]}-{bill_month[4:]}-01',
                f'{project_id}': f'{cst_row.get("project_name")}',
                'Project': cst_row.get('project_name'), 'ProjectId': project_id, 'Actual': total_cost,
                'Account': parents_folders.get('parent1', 'NA'), 'Forecast': 0, 'folder_ids': folder_ids, **parents_folders
            }
            agg_data[index]['timestamp'] = datetime.strptime(agg_data[index]['start_date'], '%Y-%m-%d')
            month = datetime.strftime(agg_data[index]['timestamp'], "%Y %b")
            agg_data[index]['Month'] = month
            agg_data[index]['filter_date'] = f'{agg_data[index]["start_date"]}-{month.split()[-1]}'
        return self.__organized_results(list(agg_data.values()))

    @logger_time_stamp
    def __get_cost_and_upload(self):
        """
        This method collect the cost and uploads to the ElasticSearch"
        :return:
        """
        collected_data = self.__get_big_query_data()
        forecast_data = self.__forecast_for_next_months(cost_data=collected_data)
        upload_data = collected_data + forecast_data
        self.elastic_upload.es_upload_data(items=upload_data, set_index='index_id')
        return upload_data

    @logger_time_stamp
    def run(self):
        """
        This method run the gcp cost explorer methods
        :return:
        """
        return self.__get_cost_and_upload()



import datetime
import os
import smtplib
from email.message import EmailMessage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import Union, Optional

from cloud_governance.common.clouds.aws.s3.s3_operations import S3Operations
from cloud_governance.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from cloud_governance.common.ldap.ldap_search import LdapSearch
from cloud_governance.common.logger.init_logger import logger

# https://github.com/redhat-performance/quads/blob/master/quads/tools/postman.py
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.common.utils.api_requests import APIRequests
from cloud_governance.main.environment_variables import environment_variables


class Postfix:
    """
    This class send mail using the postfix configuration
    install postfix and configure on machine
    $ dnf install postfix -y
    $ vi /etc/postfix/main.cf
    relayhost = [smtp.corp.redhat.com]
    inet_interfaces = all
    inet_protocols = all
    relay_domains = rdu2.scalelab.redhat.com, redhat.com
    $ systemctl restart postfix
    """

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__LDAP_HOST_NAME = self.__environment_variables_dict.get('LDAP_HOST_NAME')
        self.__ldap_search = LdapSearch(ldap_host_name=self.__LDAP_HOST_NAME)
        self.reply_to = self.__environment_variables_dict.get('REPLY_TO', 'dev-null@redhat.com')
        self.__es_host = self.__environment_variables_dict.get('es_host', '')
        self.__policy = self.__environment_variables_dict.get('policy', '')
        self.__es_port = self.__environment_variables_dict.get('es_port', '')
        self.__account = self.__environment_variables_dict.get('account', '')
        self.__policy_output = self.__environment_variables_dict.get('policy_output', '')
        self.__default_admins = self.__environment_variables_dict.get('DEFAULT_ADMINS')
        self.__email_alert = self.__environment_variables_dict.get('EMAIL_ALERT')
        self.__mail_to = self.__environment_variables_dict.get('EMAIL_TO')
        self.__mail_cc = self.__environment_variables_dict.get('EMAIL_CC')
        self.__POSTFIX_HOST = environment_variables.POSTFIX_HOST
        self.__POSTFIX_PORT = environment_variables.POSTFIX_PORT
        self.bucket_name, self.key = self.get_bucket_name()
        self.__perf_services_url = os.environ.get('PERF_SERVICES_URL')
        self.__es_index = 'cloud-governance-mail-messages'
        self.__api_request = APIRequests()
        if self.__es_host:
            self.__es_operations = ElasticSearchOperations(es_host=self.__es_host, es_port=self.__es_port)
        if self.__policy_output:
            self.__s3_operations = S3Operations(region_name='us-east-1')

    def get_bucket_name(self):
        key = 'logs'
        if 's3' in self.__policy_output.lower():
            targets = self.__policy_output.split('/')
            bucket_name = targets[2]
            key = targets[3]
        else:
            bucket_name = self.__policy_output
        return bucket_name, key

    def split_emails(self, email_str: str):
        """
        This method split the mail list
        :param email_str:
        :type email_str:
        :return:
        :rtype:
        """
        splitter = "," if "," in email_str else " "
        return [item.strip() for item in email_str.split(splitter)]

    def prettify_to(self, to: Union[str, list]):
        """
        This method prettifies to
        :param to:
        :type to:
        :return:
        :rtype:
        """

        if isinstance(to, str):
            to = self.split_emails(email_str=to)
        if isinstance(to, list):
            return ','.join([item if '@redhat.com' in item else f'{item}@redhat.com' for item in to])
        return to

    def prettify_cc(self, cc: Union[str, list], to: str = ''):
        """
        This method prettifies cc
        :param to:
        :type to:
        :param cc:
        :type cc:
        :return:
        :rtype:
        """
        if isinstance(cc, str):
            cc = self.split_emails(email_str=cc)
        cc_unique_values = [cc_user if '@redhat.com' in cc_user else f'{cc_user}@redhat.com' for cc_user in cc]
        to = self.prettify_to(to=to).split(',')
        return ','.join(list(set(cc_unique_values) - set(to)))

    def __save_to_elastic(self, **kwargs):
        """
        This method uploads data to elasticsearch
        :param kwargs:
        :return:
        """
        es_data = kwargs.get('es_data')
        data = {'Policy': self.__policy,
                'Account': self.__account.upper(),
                'MessageType': kwargs.get('message_type', 'alert')}
        if es_data:
            data.update(es_data)
        if kwargs.get('resource_id'):
            data['resource_id'] = kwargs['resource_id']
        if kwargs.get('extra_purse'):
            data['extra_purse'] = round(kwargs['extra_purse'], 3)
        if kwargs.get('remaining_budget'):
            data['remaining_budget'] = kwargs['remaining_budget']
        if self.__es_host:
            self.__es_operations.upload_to_elasticsearch(data=data, index=self.__es_index)
            logger.info(f'Uploaded to es index: {self.__es_index}')
        else:
            logger.info('Error missing the es_host')

    def __save_to_s3(self, content: str, **kwargs):
        """
        This method save mail message to s3 bucket
        :return:
        """
        if kwargs.get('filename'):
            file_name = kwargs['filename'].split('/')[-1]
            date_key = datetime.datetime.now().strftime("%Y%m%d%H")
            if self.__policy_output:
                self.__s3_operations.upload_file(file_name_path=kwargs['filename'],
                                                 bucket=self.bucket_name,
                                                 key=f'{self.key}/{self.__policy}/{date_key}',
                                                 upload_file=file_name)
                s3_path = f'{self.__policy_output}/logs/{self.__policy}/{date_key}/{file_name}'
                content += f'\n\nresource_file_path: s3://{s3_path}\n\n'
                logger.info("File Saved to S3")
        self.__save_to_elastic(**kwargs)

    @logger_time_stamp
    def send_email_postfix(self, subject: str, to: any, cc: list, content: str, **kwargs):
        if self.__email_alert:
            if self.__mail_to:
                to = self.__mail_to
            if self.__mail_cc:
                cc = self.__mail_cc
            if not self.__ldap_search.get_user_details(user_name=to):
                cc.extend(self.__default_admins)
            response = {'ok': True}
            to = self.prettify_to(to)
            cc = self.prettify_cc(cc)
            kwargs.setdefault('es_data', {}).update({'To': to, 'Cc': cc, 'Message': content})
            if self.__perf_services_url:
                body = {
                    "cc": cc,
                    "to": to,
                    "subject": subject,
                    "mail_body": content,
                }
                if kwargs.get('filename'):
                    body['file_content'] = open(kwargs['filename']).read()
                    body['filename'] = kwargs['filename'].split('/')[-1]
                api_url = self.__perf_services_url + f"/postfix/send_mail"
                response = self.__api_request.post(api_url, json=body)
                response = response.json()
            else:
                msg = MIMEMultipart('alternative')
                msg["Subject"] = subject
                msg["From"] = "%s <%s>" % (
                    'cloud-governance',
                    "@".join(["noreply-cloud-governance", 'redhat.com']),
                )
                msg["To"] = to
                msg["Cc"] = cc
                # msg.add_header("Reply-To", self.reply_to)
                # msg.add_header("User-Agent", self.reply_to)
                if kwargs.get('filename'):
                    if os.path.exists(kwargs['filename']):
                        attachment = MIMEText(open(kwargs['filename']).read())
                        attachment.add_header('Content-Disposition', 'attachment',
                                              filename=kwargs['filename'].split('/')[-1])
                        msg.attach(attachment)
                if kwargs.get('mime_type'):
                    msg.attach(MIMEText(content, kwargs.get('mime_type')))
                else:
                    msg.attach(MIMEText(content))
                try:
                    with smtplib.SMTP(self.__POSTFIX_HOST, self.__POSTFIX_PORT) as s:
                        try:
                            s.send_message(msg)
                        except smtplib.SMTPException as ex:
                            logger.error(f'Error while sending mail, {ex}')
                    response = {
                        'ok': True
                    }
                except Exception as err:
                    logger.error(f'Some error occurred, {err}')
            if isinstance(response, dict) and response.get('ok'):
                logger.info(f'Mail sent successfully to {to}')
                self.__save_to_s3(content, **kwargs)
                return True
            return False



from cloud_governance.common.utils.configs import HOURS_IN_MONTH
from cloud_governance.policy.helpers.aws.aws_policy_operations import AWSPolicyOperations
from cloud_governance.common.utils.utils import Utils


class UnattachedVolume(AWSPolicyOperations):
    RESOURCE_ACTION = "Delete"

    def __init__(self):
        super().__init__()

    def run_policy_operations(self):
        """
        This method returns the list of unattached volumes
        :return:
        :rtype:
        """
        unattached_volumes = []
        available_volumes = self._get_all_volumes()
        active_cluster_ids = self._get_active_cluster_ids()
        for volume in available_volumes:
            tags = volume.get('Tags', [])
            resource_id = volume.get('VolumeId')
            cleanup_result = False
            cluster_tag = self._get_cluster_tag(tags=volume.get('Tags'))
            if Utils.equal_ignore_case(volume.get('State'), 'available') and \
                    cluster_tag not in active_cluster_ids and \
                    self.get_skip_policy_value(tags=tags) not in ('NOTDELETE', 'SKIP'):
                cleanup_days = self.get_clean_up_days_count(tags=tags)
                cleanup_result = self.verify_and_delete_resource(resource_id=resource_id, tags=tags,
                                                                 clean_up_days=cleanup_days)
                monthly_price = self._resource_pricing.get_ebs_unit_price(region_name=self._region,
                                                                          ebs_type=volume.get('VolumeType', ''))
                unit_price = (monthly_price / HOURS_IN_MONTH) * float(volume.get('Size'))
                resource_data = self._get_es_schema(resource_id=resource_id,
                                                    user=self.get_tag_name_from_tags(tags=tags, tag_name='User'),
                                                    skip_policy=self.get_skip_policy_value(tags=tags),
                                                    cleanup_days=cleanup_days, dry_run=self._dry_run,
                                                    name=self.get_tag_name_from_tags(tags=tags, tag_name='Name'),
                                                    region=self._region,
                                                    cleanup_result=str(cleanup_result),
                                                    resource_action=self.RESOURCE_ACTION,
                                                    cloud_name=self._cloud_name,
                                                    resource_type=volume.get('VolumeType', ''),
                                                    unit_price=unit_price,
                                                    resource_state=volume.get(
                                                        'State') if not cleanup_result else "Deleted",
                                                    volume_size=f"{volume.get('Size')} GB"
                                                    )
                unattached_volumes.append(resource_data)
                if not cleanup_result:
                    self.update_resource_tags(resource_id=resource_id, tags=tags + self.cost_savings_tag)
            else:
                cleanup_days = 0
                self.delete_resource_tags(resource_id=resource_id, tags=self.cost_savings_tag)
            if not cleanup_result:
                self.update_resource_day_count_tag(resource_id=resource_id, cleanup_days=cleanup_days, tags=tags)

        return unattached_volumes



from datetime import datetime

import boto3

from cloud_governance.cloud_resource_orchestration.clouds.aws.ec2.collect_cro_reports import CollectCROReports
from cloud_governance.cloud_resource_orchestration.clouds.aws.ec2.cost_over_usage import CostOverUsage
from cloud_governance.cloud_resource_orchestration.clouds.aws.ec2.monitor_cro_instances import MonitorCROInstances
from cloud_governance.cloud_resource_orchestration.clouds.aws.ec2.aws_monitor_tickets import AWSMonitorTickets
from cloud_governance.cloud_resource_orchestration.clouds.aws.ec2.tag_cro_instances import TagCROInstances
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.main.environment_variables import environment_variables


class RunCRO:

    PERSISTENT_RUN_DOC_ID = f'cro_run_persistence-{datetime.utcnow().date()}'
    PERSISTENT_RUN_INDEX = 'cloud_resource_orchestration_persistence_run'

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.cro_cost_over_usage = CostOverUsage()
        self.cro_reports = CollectCROReports()
        self.aws_monitor_tickets = AWSMonitorTickets()
        self.account = self.__environment_variables_dict.get('account', '').replace('OPENSHIFT-', '').lower().strip()
        self.__run_active_regions = self.__environment_variables_dict.get('RUN_ACTIVE_REGIONS')
        self.__region = self.__environment_variables_dict.get('AWS_DEFAULT_REGION', '')

    @logger_time_stamp
    def send_cro_alerts(self):
        """
        This method send the cost_over_usage alert and Ticket status alerts
        :return:
        """
        es_data = self.cro_cost_over_usage.es_operations.get_es_data_by_id(index=self.PERSISTENT_RUN_INDEX, id=self.PERSISTENT_RUN_DOC_ID)
        first_run = True
        if es_data:
            source = es_data.get('_source')
            last_run_time = source.get(f'last_run_{self.account.lower()}')
            if last_run_time:
                last_updated_time = datetime.strptime(last_run_time, "%Y-%m-%dT%H:%M:%S.%f").date()
                if last_updated_time == datetime.utcnow().date():
                    first_run = False
        self.__environment_variables_dict.update({'CRO_FIRST_RUN': first_run})
        if first_run:
            cost_over_usage_users = self.cro_cost_over_usage.run()
            logger.info(f'Cost Over Usage Users list: {", ".join(cost_over_usage_users)}')
            self.aws_monitor_tickets.run()
            self.cro_reports.update_in_progress_ticket_cost()
        self.save_current_timestamp()

    @logger_time_stamp
    def save_current_timestamp(self):
        """
        This method saves the current timestamp
        Storing timestamp for not sending multiple alerts in a day, if we run any number of times
        :return:
        """
        if not self.cro_cost_over_usage.es_operations.verify_elastic_index_doc_id(index=self.PERSISTENT_RUN_INDEX, doc_id=self.PERSISTENT_RUN_DOC_ID):
            self.cro_cost_over_usage.es_operations.upload_to_elasticsearch(index=self.PERSISTENT_RUN_INDEX, data={f'last_run_{self.account}': datetime.utcnow()}, id=self.PERSISTENT_RUN_DOC_ID)
        else:
            self.cro_cost_over_usage.es_operations.update_elasticsearch_index(index=self.PERSISTENT_RUN_INDEX, metadata={f'last_run_{self.account}': datetime.utcnow()}, id=self.PERSISTENT_RUN_DOC_ID)

    @logger_time_stamp
    def run_cloud_resources(self):
        """
        This method run the aws resources in specified region or all regions
        :return:
        """
        if self.__run_active_regions:
            active_regions = [region.get('RegionName') for region in boto3.client('ec2').describe_regions()['Regions']]
            logger.info(f"""*****Running CloudResourceOrchestration in all Active regions: {active_regions}*****""")
        else:
            active_regions = [self.__region]
        for active_region in active_regions:
            cro_monitor = MonitorCROInstances(region_name=active_region)
            cro_tagging = TagCROInstances(region_name=active_region)
            self.__environment_variables_dict.update({'AWS_DEFAULT_REGION': active_region})
            logger.info(f"""Running CloudResourceOrchestration in region: {active_region}""")
            logger.info(f"""{active_region}: -> Running CRO Tagging""")
            tagging_response = cro_tagging.run()
            logger.info(f'Tagged instances : {tagging_response}')
            logger.info(f"""{active_region}: -> Running CRO Resource data Collection""")
            monitor_response = cro_monitor.run()
            if monitor_response:
                cro_reports = self.cro_reports.run(monitor_response)
                logger.info(f'Cloud Orchestration Resources: {cro_reports}')

    @logger_time_stamp
    def start_cro(self):
        """
        This method start the cro process methods
        1. Send alert to cost over usage users
        2. Tag the new instances
        3. monitor and upload the new instances' data
        4. Monitor the Jira ticket progressing
        :return:
        """
        self.send_cro_alerts()
        self.run_cloud_resources()

    @logger_time_stamp
    def run(self):
        """
        This method start the aws CRO operations
        :return:
        """
        self.start_cro()



import os

from cloud_governance.common.clouds.cloudability.cloudability_operations import CloudabilityOperations
from cloud_governance.common.clouds.cloudability.templates.cloudability_dimensions import COST_USAGE_REPORT_DIMENSIONS
from cloud_governance.common.elasticsearch.elastic_upload import ElasticUpload
from cloud_governance.common.elasticsearch.modals.cost_usage_reports_data import CostUsageReportData


class CostUsageReports:
    REPORT_GENERATE_TYPE = 'cloudability'

    def __init__(self):
        self.__cloudability_operations = CloudabilityOperations()
        self.__account_id = self.__cloudability_operations.environment_variables_dict.get("IBM_ACCOUNT_ID")
        self.elastic_upload = ElasticUpload()

    def __generate_es_schemed_data(self, cloudability_reports: list):
        """
        This method returns the report data that satisfy the ES Data Scheme
        :return:
        """
        modified_cloudability_reports = []
        for item in cloudability_reports:
            modified_cloudability_reports.append(
                CostUsageReportData(report_generated_type=self.REPORT_GENERATE_TYPE, **item).to_dict())
        return modified_cloudability_reports

    def collect_reports(self):
        """
        This method collects reports and publishes to elasticsearch
        :return:
        """

        custom_filter = f'&filters=vendor_account_identifier=={self.__account_id}'
        dimensions = ','.join(COST_USAGE_REPORT_DIMENSIONS.keys())
        cloudability_reports = self.__cloudability_operations.get_cost_reports(custom_filter=custom_filter,
                                                                               dimensions=dimensions,
                                                                               look_back_days=1)
        if cloudability_reports:
            account = cloudability_reports[0]['vendor_account_name']
            self.__cloudability_operations.environment_variables_dict['account'] = account
            os.environ['account'] = account
        cloudability_reports = self.__generate_es_schemed_data(cloudability_reports)
        return cloudability_reports

    def run(self):
        """
        This method runs the tag operations
        :return:
        """
        return self.collect_reports()



from functools import wraps

from cloud_governance.common.clouds.ibm.classic.classic_operations import ClassicOperations
from cloud_governance.common.clouds.ibm.developer_tools.schematic_operations import SchematicOperations
from cloud_governance.common.clouds.ibm.platform_services.platform_service_operations import PlatformServiceOperations
from cloud_governance.common.clouds.ibm.tagging.global_tagging_operations import GlobalTaggingOperations
from cloud_governance.common.clouds.ibm.vpc.vpc_infra_operations import VpcInfraOperations
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp


def get_resources_wrapper(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        resources_crn = []
        resource_list = func(*args, **kwargs)
        for region, resources in resource_list.items():
            resources_crn.extend(
                [resource.get('crn') if isinstance(resource, dict) else resource for resource in resources])
        return list(set(resources_crn))

    return wrapper


class TagResources:
    """
    This class tags the Virtual PrivateCloud Resources
    Virtual Servers
    VPC Resources
    """

    def __init__(self):
        self.vpc_infra_operations = VpcInfraOperations()
        self.tag_operations = GlobalTaggingOperations()
        self.schematic_operations = SchematicOperations()
        self._classic_operations = ClassicOperations()
        self.platform_operations = PlatformServiceOperations()
        self.__env_config = self.vpc_infra_operations.env_config
        self.__ibm_custom_tags_list = self.__env_config.IBM_CUSTOM_TAGS_LIST \
            if hasattr(self.__env_config, 'IBM_CUSTOM_TAGS_LIST') \
            else self.__env_config.environment_variables_dict.get('IBM_CUSTOM_TAGS_LIST')
        self.resource_to_tag = self.__env_config.RESOURCE_TO_TAG \
            if hasattr(self.__env_config, 'RESOURCE_TO_TAG') \
            else self.__env_config.environment_variables_dict.get('RESOURCE_TO_TAG')

    @get_resources_wrapper
    @logger_time_stamp
    def get_virtual_servers_crn(self):
        """
        This method returns all virtual server crn's
        :return:
        """
        return self.vpc_infra_operations.get_all_instances()

    @get_resources_wrapper
    @logger_time_stamp
    def get_schematics_workspaces_crn(self):
        """
        This method returns all schematics workspaces crn's
        :return:
        """
        return self.schematic_operations.get_all_workspaces()

    @get_resources_wrapper
    @logger_time_stamp
    def get_resource_instances_crn(self):
        """
        This method returns all resource instances crn's
        :return:
        """
        return self.platform_operations.get_resource_instances()

    @logger_time_stamp
    def get_classic_baremetals_crn(self):
        """
        This method returns all classic baremetals crn's
        :return:
        """
        hardware_ids = self._classic_operations.get_hardware_ids()
        return hardware_ids

    @logger_time_stamp
    def get_classic_virtual_machines_crn(self):
        """
        This method returns all classic baremetals crn's
        :return:
        """
        virtual_machine_ids = self._classic_operations.get_virtual_machine_ids()
        return virtual_machine_ids

    @logger_time_stamp
    def tag_all_resources(self):
        """
        This method tags all Virtual PrivateCloud Resources
        :return:
        """
        if not self.__ibm_custom_tags_list:
            return {'ok': False, 'errors': {},
                    'message': 'No tags to add resources, please export IBM_CUSTOM_TAGS_LIST in '
                               'str format. i.e "key:value, env:test"'}
        tags_list = self.__ibm_custom_tags_list.split(',')
        vpc_resources = [
            "resource_instances",
            "virtual_servers",
            "schematics_workspaces",
            "classic_baremetals",
            "classic_virtual_machines"
        ]
        if self.resource_to_tag:
            resources_to_tag = [resource.strip() for resource in self.resource_to_tag.split(',')]
            vpc_resources = list(set(vpc_resources) & set(resources_to_tag))
        logger.info(f"Running tag operation on total of {len(vpc_resources)} resources")
        errors = messages = {}
        for vpc_resource in vpc_resources:
            resources_crn = getattr(self, f'get_{vpc_resource}_crn')()
            if vpc_resource == 'classic_baremetals':
                for resource in resources_crn:
                    ok, errors_list = self._classic_operations. \
                        update_baremetal_tags(tags=tags_list, hardware_id=resource.get('id'))
                    self.process_ok_and_errors(ok, errors_list, messages, vpc_resource, errors)
            elif vpc_resource == 'classic_virtual_machines':
                for resource in resources_crn:
                    ok, errors_list = self._classic_operations. \
                        update_virtual_machine_tags(tags=tags_list, virtual_machine_id=resource.get('id'))
                    self.process_ok_and_errors(ok, errors_list, messages, vpc_resource, errors)
            else:
                logger.info(f"Started the tagging operation for {vpc_resource}")
                ok, errors_list = self.tag_operations.update_tags(resources_crn, tags=tags_list)
                self.process_ok_and_errors(ok, errors_list, messages, vpc_resource, errors)

        return {'errors': errors, 'messages': messages}

    def process_ok_and_errors(self, ok: bool, errors_list: any, messages: dict, resource: str, errors: dict):
        """
        This method processes ok and errors
        :param errors:
        :param ok:
        :param errors_list:
        :param messages:
        :param resource:
        :return:
        """
        success_message = 'tagged are added to all resources'
        if not ok:
            message = 'Unable to tag all resources'
            logger.info(f'{message}, please find the servers that are not tagged: {errors_list}')
            errors.update({resource: message, 'crns': errors_list})
        else:
            messages.update({resource: success_message})

    def run(self):
        """
        This method runs the tag operations
        :return:
        """
        return self.tag_all_resources()



import os.path
import tempfile

import numpy as np
import pandas as pd

from cloud_governance.common.google_drive.google_drive_operations import GoogleDriveOperations
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.main.environment_variables import environment_variables


class UploadToGsheet:
    """
    This class collects the data from clouds and uploads to the GSheet
    """

    def __init__(self):
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.gsheet_operations = GoogleDriveOperations()
        self.__gsheet_id = self.__environment_variables_dict.get('SPREADSHEET_ID', '')

    @logger_time_stamp
    def get_cost_center_budget_details(self, account_id: str, dir_path: str = ''):
        """
        This method returns the cost center & budget details
        @return:
        """
        with tempfile.TemporaryDirectory() as tmp_dir:
            sheet_name = 'Accounts'
            dirtectory = dir_path if dir_path else tmp_dir
            file_path = f'{tmp_dir}/{sheet_name}.csv' if not dir_path else f'{dir_path}/{sheet_name}.csv'
            if not os.path.exists(file_path):
                self.gsheet_operations.download_spreadsheet(spreadsheet_id=self.__gsheet_id, sheet_name=sheet_name,
                                                            file_path=dirtectory)
            accounts_df = pd.read_csv(file_path)
            account_row = accounts_df[
                accounts_df['AccountId'] == account_id].reset_index().to_dict(
                orient='records')
            if account_row:
                return account_row[0].get('CostCenter', 0), round(
                    float(account_row[0].get('Budget', '0').replace(',', '')), 0), str(account_row[0].get('Year')), account_row[0].get('Owner')
            return 0, 0, '', 'Others'

    def get_monthly_spa(self, month_name: str, dir_path: str = ''):
        """This method gets the monthly savings plan amortization"""
        with tempfile.TemporaryDirectory() as tmp_dir:
            sheet_name = 'ASP'
            dirtectory = dir_path if dir_path else tmp_dir
            file_path = f'{tmp_dir}/{sheet_name}.csv' if not dir_path else f'{dir_path}/{sheet_name}.csv'
            if not os.path.exists(file_path):
                self.gsheet_operations.download_spreadsheet(spreadsheet_id=self.__gsheet_id, sheet_name=sheet_name, file_path=dirtectory)
            accounts_df = pd.read_csv(file_path)
            records = accounts_df.to_dict(orient='records')
            for record in records:
                if record.get('Month').lower() == month_name.lower():
                    return float(record.get('Total'))
            return 0

    def format_for_updating_the_cells(self, update_data: list, gsheet_data: pd, sheet_name: str, doc_id: str, doc_id2: str = ''):
        """
        This method format the data to be updated to GSheet and update in the GSheet
        @param doc_id2:
        @param update_data:
        @param gsheet_data:
        @param sheet_name:
        @param doc_id:
        @return:
        """
        gsheet_index = gsheet_data.index
        for data in update_data:
            if doc_id2:
                index = np.where((gsheet_data[doc_id] == data[0]) & (gsheet_data[doc_id2] == data[2]))
            else:
                index = gsheet_data[doc_id] == data[0]
            cell_row = gsheet_index[index].tolist()
            if cell_row:
                row_data = gsheet_data.iloc[cell_row[0]].to_list()
                if set(row_data) != set(data):
                    cell_row = cell_row[0] + 1
                    upload_data = []
                    for value in data:
                        if isinstance(value, str):
                            upload_data.append({"userEnteredValue": {'stringValue': value}})
                        elif isinstance(value, int):
                            upload_data.append({"userEnteredValue": {'numberValue': value}})
                        else:
                            if isinstance(value, float):
                                upload_data.append({"userEnteredValue": {'numberValue': value}})
                    self.gsheet_operations.update_row_in_gsheet(data=upload_data, gsheet_id=self.__gsheet_id, row=cell_row, sheet_name=sheet_name)
            else:
                self.gsheet_operations.append_values(spreadsheet_id=self.__gsheet_id, sheet_name=sheet_name, values=[data])

    def update_cloud_agg_data(self, cloud_name: str, cloud_alias_name: str):
        """
        This method aggregate all the cloud data and updates if there is any difference
        @param cloud_alias_name:
        @param cloud_name:
        @return:
        """
        self.gsheet_operations.create_work_sheet(gsheet_id=self.__gsheet_id, sheet_name=cloud_name)
        with tempfile.TemporaryDirectory() as tmp_dir:
            file_path = os.path.join(tmp_dir, f'{cloud_alias_name}.csv')
            if not os.path.exists(file_path):
                self.gsheet_operations.download_spreadsheet(spreadsheet_id=self.__gsheet_id, sheet_name=cloud_alias_name, file_path=tmp_dir)
            cloud_data = pd.read_csv(file_path)
            account_id = cloud_data['AccountId'].tolist()[0]
            account_name = cloud_data['Account'].tolist()[0]
            cost_center = cloud_data['CostCenter'].tolist()[0]
            months_data = {}
            for index, row in cloud_data.reset_index().iterrows():
                start_year = row['Month'].split()[0]
                if start_year not in months_data:
                    months_data[start_year] = {}
                    months_data[start_year]['AccountId'] = account_id
                    months_data[start_year]['Account'] = account_name
                    months_data[start_year]['Year'] = int(start_year)
                    months_data[start_year]['CostCenter'] = cost_center
                    months_data[start_year]['ActualCost'] = round(row['ActualCost'], 3)
                    months_data[start_year]['Forecast'] = round(row['Forecast'], 3)
                    months_data[start_year]['Available'] = round(row['AllocatedBudget'] - months_data[start_year]['ActualCost'], 3)
                    months_data[start_year]['AllocatedBudget'] = row['AllocatedBudget']
                    months_data[start_year]['TotalMonths'] = 1
                else:
                    months_data[start_year]['ActualCost'] += round(row['ActualCost'], 3)
                    months_data[start_year]['Forecast'] += round(row['Forecast'], 3)
                    months_data[start_year]['Available'] = round(row['AllocatedBudget'] - months_data[start_year]['ActualCost'], 3)
                    months_data[start_year]['TotalMonths'] += 1
            data_upload_keys = ['AccountId', 'Account', 'Year', 'CostCenter', 'Actual', 'Forecast', 'Available', 'AllocatedBudget', 'TotalMonths']
            data_upload_values = [list(item.values()) for item in list(months_data.values())]
            file_path = os.path.join(tmp_dir, f'{cloud_name}.csv')
            if not os.path.exists(file_path):
                self.gsheet_operations.download_spreadsheet(spreadsheet_id=self.__gsheet_id, sheet_name=cloud_name, file_path=tmp_dir)
            if os.path.exists(file_path):
                cloud_agg_df = pd.read_csv(file_path)
                self.format_for_updating_the_cells(update_data=data_upload_values, gsheet_data=cloud_agg_df,
                                                   sheet_name=cloud_name, doc_id='AccountId', doc_id2='Year')
            else:
                with tempfile.NamedTemporaryFile(mode='w', suffix='.csv') as tmp_file:
                    tmp_file.write(f'{",".join([str(item) for item in data_upload_keys])}\n')
                    for data_values in data_upload_values:
                        tmp_file.write(f'{",".join([str(item) for item in data_values])}\n')
                    tmp_file.flush()
                    self.gsheet_operations.paste_csv_to_gsheet(csv_path=tmp_file.name, spreadsheet_id=self.__gsheet_id, sheet_name=cloud_name)

    @logger_time_stamp
    def update_data(self, cloud_data: dict):
        """
        This method update the data in the Gsheet and updates the data if there is any difference
        @return:
        """
        cloud_alias_name = cloud_data.get('cloud_alias_name').upper()
        cloud_name = cloud_data.get('cloud_name').upper()
        with tempfile.NamedTemporaryFile(suffix='.csv', mode='w') as tmp_file:
            file_path = f'/tmp/{cloud_alias_name}.csv'
            self.gsheet_operations.create_work_sheet(gsheet_id=self.__gsheet_id, sheet_name=cloud_alias_name)
            if not os.path.exists(file_path):
                self.gsheet_operations.download_spreadsheet(spreadsheet_id=self.__gsheet_id, sheet_name=cloud_alias_name, file_path=f'/tmp')
            gsheet_row_keys = ['_id', 'AccountId', 'Account', 'Month', 'CostCenter', 'ActualCost', 'Budget', 'Forecast', 'AllocatedBudget', 'AvailableBalance']
            gsheet_row_values = []
            for data in cloud_data.get('cloud_data'):
                data_values = [data.get('index_id'), data.get('AccountId'), data.get('Account'), data.get('Month'), data.get('CostCenter'),
                               round(data.get('Actual'), 3), round(data.get('Budget'), 3), round(float(data.get('Forecast')), 3),
                               int(data.get('AllocatedBudget')), round(data.get('Budget') - data.get('Actual'), 3)]
                gsheet_row_values.append(data_values)
            if os.path.exists(file_path):
                data_in_gsheet = pd.read_csv(file_path)
                self.format_for_updating_the_cells(update_data=gsheet_row_values, gsheet_data=data_in_gsheet, sheet_name=cloud_alias_name, doc_id='_id')
                os.remove(file_path)
            else:
                total_data = [gsheet_row_keys]
                total_data.extend(gsheet_row_values)
                for data in total_data:
                    tmp_file.write(f'{",".join((str(item) for item in data))}\n')
                tmp_file.flush()
                self.gsheet_operations.paste_csv_to_gsheet(csv_path=tmp_file.name, spreadsheet_id=self.__gsheet_id, sheet_name=cloud_alias_name)
        self.update_cloud_agg_data(cloud_name=cloud_name, cloud_alias_name=cloud_alias_name)



LOOK_BACK_DAYS = 30
MONTHS = 12
DEFAULT_ROUND_DIGITS = 3
PUBLIC_ACCOUNTS_COST_REPORTS = 'Accounts'
IT_ACCOUNTS_COST_REPORTS_LIST = 'IT_ACCOUNTS'
HOURS_IN_SECONDS = 3600
HOURS_IN_DAY = 24
HOURS_IN_MONTH = 720
TOTAL_BYTES_IN_KIB = 1024
INSTANCE_START_PREFIX = 'i-'

DATE_FORMAT = "%Y-%m-%d"
DATE_TIME_FORMAT_T = "%Y-%m-%dT%h:%m"
UNUSED_DAYS = 7

# Policy Values
# Instance Idle
INSTANCE_IDLE_DAYS = 7
INSTANCE_IDLE_CPU_PERCENTAGE = 2
INSTANCE_IDLE_NETWORK_IN_KILO_BYTES = 5  # In KiB
INSTANCE_IDLE_NETWORK_OUT_KILO_BYTES = 5  # In KiB
EC2_NAMESPACE = 'AWS/EC2'
CLOUDWATCH_METRICS_AVAILABLE_DAYS = 14
AWS_DEFAULT_GLOBAL_REGION = 'us-east-1'
UNUSED_ACCESS_KEY_DAYS = 90
UNUSED_ACCESS_KEY_MAX_DAY = 1000

# X86 to Graviton
GRAVITON_MAPPINGS = {
    'c6a': 'c6g',
    'c6i': 'c6g',
    'c6in': 'c6gn',
    'c7': 'c7g',
    'c7i': 'c7g',
    'c7a': 'c7g',
    'c7i-flex': 'c7g',
    'g5': 'g5g',
    'hpc7a': 'hpc7g',
    'i4i': 'i4g',
    'm5': 'm6g',
    'm5a': 'm6g',
    'm5n': 'm6g',
    'm5zn': 'm6g',
    'm6i': 'm6g',
    'm6in': 'm6g',
    'm6a': 'm6g',
    'm7i': 'm7g',
    'm7a': 'm7g',
    'm7i-flex': 'm7g',
    'r7i': 'r8g',
    'r7a': 'r7g',
    'r5': 'r7g',
    'r6i': 'r6g',
    't3': 't4g',
    't2': 'a1'
}

DEFAULT_GRAVITON_INSTANCE = 'm6g'
